
from __future__ import annotations

import functools
import json
import types
from typing import *

from aiohttp.web import HTTPException, Request, Response
from jinja2 import Environment, FileSystemLoader, Template, select_autoescape
from jinja2.exceptions import (TemplateAssertionError, TemplateError,
                               TemplateNotFound, TemplateSyntaxError,
                               UndefinedError)
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

from .logging import setup_logging

logger = setup_logging(__name__)

Res = TypeVar("Res", bound=Response)
Req = TypeVar("Req", bound=Request)
Exc = TypeVar("Exc", bound=HTTPException)
View = Callable[[Req], Awaitable[Res]]
P = ParamSpec("P")
V = TypeVar("V", bound=View)

class HTMLResponse(Response):
    def __init__(self, text: str, status: int = 200, **kwargs: Any) -> None:
        super().__init__(text=text, status=status, **kwargs)
        self.headers["Content-Type"] = "text/html"
        self.content_type = "text/html"
        
class JSONResponse(Response):
    def __init__(self, data: Any, status: int = 200, **kwargs: Any) -> None:
        super().__init__(text=json.dumps(data, indent=4), status=status, **kwargs)
        self.headers["Content-Type"] = "application/json"
        self.content_type = "application/json"
        
class PlainTextResponse(Response):
    def __init__(self, text: str, status: int = 200, **kwargs: Any) -> None:
        super().__init__(text=text, status=status, **kwargs)
        self.headers["Content-Type"] = "text/plain"
        self.content_type = "text/plain"

jinja_env = Environment(
    loader=FileSystemLoader("templates"),
    autoescape=select_autoescape(["html", "xml"]),
    enable_async=True,
)    

async def render_template(
        template_name: str, context:Dict[str,Any]) -> HTMLResponse:
    """
    Renders a template and returns a response.
    """
    try:
        template_ = jinja_env.get_template(template_name)
        html = await template_.render_async(**context)
        return HTMLResponse(html)
    except (TemplateError, TemplateAssertionError, TemplateSyntaxError, TemplateNotFound, UndefinedError) as e:
        logger.error(e)
        raise HTTPException(reason=str(e)) from e
    
async def render_template_string(
        template_string: str, context:Dict[str,Any]) -> HTMLResponse:
    """
    Renders a template and returns a response.
    """
    try:
        template_ = Template(template_string)
        html = await template_.render_async(**context)
        return HTMLResponse(html)
    except (TemplateError, TemplateAssertionError, TemplateSyntaxError, TemplateNotFound, UndefinedError) as e:
        logger.error(e)
        raise HTTPException(reason=str(e)) from e
    
    
def template(template_name:str, func:Callable[[Request],Awaitable[Dict[str,Any]]]) -> Callable[[Request], Awaitable[HTMLResponse]]:
    """
    Decorator to render a template and return a response.
    """
    @functools.wraps(func)
    async def wrapper(request:Request) -> HTMLResponse:
        try:
            template_ = jinja_env.get_template(template_name)
            context = await func(request)
            html = await template_.render_async(**context)
            return HTMLResponse(html)
        except (TemplateError, TemplateAssertionError, TemplateSyntaxError, TemplateNotFound, UndefinedError) as e:
            logger.error(e)
            raise HTTPException(reason=str(e)) from e
    return wrapper






import json
from abc import ABC, abstractmethod
from typing import (Any, Dict, Generic, Iterable, List, Optional, Type,
                    TypeVar, Union, cast)

import openai
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

from .api.openai import Message, Model
from .logging import log

Vector = List[float]
MetaData = Dict[str, str]

T = TypeVar("T")

class LazyProxy(Generic[T], ABC):
    def __init__(self) -> None:
        self.__proxied: Union[T, None] = None

    def __getattr__(self, attr: str) -> object:
        return getattr(self.__get_proxied__(), attr)

    def __repr__(self) -> str:
        return repr(self.__get_proxied__())

    def __dir__(self) -> Iterable[str]:
        return self.__get_proxied__().__dir__()

    def __get_proxied__(self) -> T:
        proxied = self.__proxied
        if proxied is not None:
            return proxied

        self.__proxied = proxied = self.__load__()
        return proxied

    def __set_proxied__(self, value: T) -> None:
        self.__proxied = value

    def __as_proxied__(self) -> T:
        """Helper method that returns the current proxy, typed as the loaded object"""
        return cast(T, self)

    @abstractmethod
    def __load__(self) -> T:
        ...

class FunctionRequest(BaseModel):
    """
    Defines a function request.
    """

    model: Model = Field(
        default="gpt-4-0613", description="The model used for the chat completion."
    )
    messages: List[Message] = Field(
        ..., description="The list of messages in the conversation."
    )
    functions: Optional[List[Dict[str, Any]]] = Field(
        None, description="Optional list of functions to be used."
    )


@log
class FunctionType(BaseModel, ABC):
    _subclasses: List[Type["FunctionType"]] = []

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _schema = cls.schema()
        if cls.__doc__ is None:
            raise ValueError(
                f"FunctionType subclass {cls.__name__} must have a docstring"
            )
        cls.openaischema = {
            "name": cls.__name__,
            "description": cls.__doc__,
            "parameters": {
                "type": "object",
                "properties": _schema["properties"],
                "required": _schema["required"],
            },
        }
        cls.logger.info(f"\tRegistered function {cls.__name__}")  # type: ignore # pylint: disable=no-member
        cls._subclasses.append(cls)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.logger.info(f"Function {self.__class__.__name__} called with {kwargs}")  # type: ignore # pylint: disable=no-member
        self.logger.info(f"Function {self.__class__.__name__} initialized")  # type: ignore # pylint: disable=no-member

    @abstractmethod
    async def run(self) -> Any:
        ...


F = TypeVar("F", bound=FunctionType)


async def parse_openai_response(  # pylint: disable=dangerous-default-value
    response: dict,
    functions: List[
        Type[F]
    ] = FunctionType._subclasses,  # pylint: disable=protected-access
) -> Any:
    """Parse the response from OpenAI and return the result."""
    choice = response["choices"][0]["message"]
    if "function_call" in choice:
        function_call_ = choice["function_call"]
        name = function_call_["name"]
        arguments = function_call_["arguments"]
        for i in functions:
            if i.__name__ == name:
                result = await i.run(i(**json.loads(arguments)))
                break
        else:
            raise ValueError(f"Function {name} not found")
        return result
    return choice["content"]


async def function_call(  # pylint: disable=dangerous-default-value
    text: str,
    context: Optional[str] = None,
    model: Model = "gpt-4-0613",
    functions: List[
        Type[F]
    ] = FunctionType._subclasses,  # pylint: disable=protected-access
) -> Any:
    """
    Function to call a OpenAI function with given text and context.

    Arguments:
    text -- Input text for the function
    context -- Optional context for the function
    model -- Model to be used. Defaults to "gpt-4-0613"
    functions -- List of function types. Defaults to all subclasses of FunctionType.
    """
    if context is None:
        messages = [Message(role="user", content=text)]
    else:
        messages = [
            Message(role="system", content=context),
            Message(role="user", content=text),
        ]
    data = FunctionRequest(
        model=model, messages=messages, functions=[i.openaischema for i in functions]
    ).dict()
    response = await openai.ChatCompletion.acreate(**data)
    assert isinstance(response, dict)
    return await parse_openai_response(response, functions=functions)



"""Logging and error handling utilities for the OpenAI Function Python package."""
import logging
from typing import Any, Callable, Coroutine, TypeVar

from aiohttp.web_exceptions import HTTPException
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import install
from rich.traceback import install as ins

# Typing the return type of the wrapped function.
T = TypeVar("T")


def setup_logging(name: str) -> logging.Logger:
    """
    Set's up logging using the Rich library for pretty and informative terminal logs.

    Arguments:
    name -- Name for the logger instance. It's best practice to use the name of the module where logger is defined. # pylint: disable=line-too-long
    """

    # Install pretty representations of data structures using Rich library.
    install()

    # Install Rich traceback handler.
    ins()

    # Create a Console object that can record terminal output.
    console = Console(record=True, force_terminal=True)

    # Create a RichHandler for rich logging.
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        tracebacks_extra_lines=5,
        tracebacks_theme="solarized",
    )

    # Set the log level and formatter for the console handler.
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    console_handler.setLevel(logging.INFO)

    # Setup basic configuration for logging.
    logging.basicConfig(level=logging.INFO, handlers=[console_handler])

    # Create and return a logger with the specified name.
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)

    return logger_


# Set up a logger for this module.
logger = setup_logging(__name__)


def handle_errors(
    func: Callable[..., Coroutine[Any, Any, T]]
) -> Callable[..., Coroutine[Any, Any, T]]:  # pylint: disable=line-too-long
    """
    A decorator to handle errors in an asynchronous function.

    Arguments:
    func -- The asynchronous function whose errors are to be handled.
    """

    async def wrapper(*args: Any, **kwargs: Any) -> T:
        """
        Wrapper function to handle errors in the function call.
        """
        try:
            logger.info("Calling %s", func.__name__)
            return await func(*args, **kwargs)
        except HTTPException as exc:
            logger.error(exc.__class__.__name__)
            logger.error(exc.reason)
            raise exc from exc
        except Exception as exc:
            logger.error(exc.__class__.__name__)
            logger.error(str(exc))
            raise exc from exc

    return wrapper




def log(cls):
    """Adds a logger to a class."""
    if not hasattr(cls, "logger"):
        cls.logger = setup_logging(cls.__name__)
    return cls



"""REST API Module with automatic OpenAPI generation."""
import asyncio
import os
import threading
import time
from functools import wraps
from inspect import signature
from typing import Awaitable, Callable

from aiohttp.typedefs import Handler
from aiohttp.web import (Application, AppRunner, FileResponse, Request,
                         Response, StreamResponse, TCPSite, json_response,
                         run_app)
from aiohttp.web_exceptions import HTTPException
from aiohttp.web_middlewares import middleware
from aiohttp.web_ws import WebSocketResponse
from aiohttp_sse import EventSourceResponse, sse_response

from .client import APIClient, FaunaClient
from .docs import extract, html, load, transform
from .helpers import do_response
from .json import jsonable_encoder
from .logging import setup_logging
from .odm import FaunaModel

Middleware = Callable[[Request, Handler], Awaitable[StreamResponse]]


class APIServer(Application):
    """Aiohttp Application with automatic OpenAPI generation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, logger=setup_logging(self.__class__.__name__), **kwargs)
        self.openapi = {
            "openapi": "3.0.0",
            "info": {"title": "AioFauna", "version": "1.0.0"},
            "paths": {},
            "components": {"schemas": {}},
            "description": "AioFauna API",
        }
        self._route_open_api_params = {}

        @self.get("/openapi.json")
        async def openapi():
            response = jsonable_encoder(self.openapi)
            return response

        @self.get("/docs")
        async def docs():
            return Response(text=html, content_type="text/html")
        
        @self.on_event("startup")
        async def startup(_):
            await FaunaModel.create_all()
            
        @self.on_event("shutdown")
        async def shutdown(_):
            await APIClient.cleanup()
            await FaunaModel.cleanup()
    def document(self, path: str, method: str):
        """

        Decorator to document a function.

        """

        def decorator(func):
            sig = signature(func)
            params = sig.parameters
            open_api_params = extract(params.copy(), path)
            self._route_open_api_params[(path, method)] = open_api_params
            transform(self.openapi, path, method, func, open_api_params)

            async def wrapper(*args, **kwargs):
                request: Request = args[0]
                args = args[1:]
                args_to_apply = await load(request, params.copy())
                definitive_args = {}
                for name, param in params.items():
                    if name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                if asyncio.iscoroutinefunction(func):
                    response = await func(*args, **kwargs, **definitive_args)
                else:
                    response = func(*args, **kwargs, **definitive_args)
                return do_response(response)

            func.injectable = True
            wrapper._handler = func
            return wrapper

        return decorator

    def get(self, path: str, **kwargs):
        """GET decorator"""

        def decorator(func):
            self.router.add_get(path, self.document(path, "GET")(func), **kwargs)
            return func

        return decorator

    def post(self, path: str, **kwargs):
        """POST decorator"""

        def decorator(func):
            self.router.add_post(path, self.document(path, "POST")(func), **kwargs)
            return func

        return decorator

    def put(self, path: str, **kwargs):
        """PUT decorator"""

        def decorator(func):
            self.router.add_put(path, self.document(path, "PUT")(func), **kwargs)
            return func

        return decorator

    def delete(self, path: str, **kwargs):
        """DELETE decorator"""

        def decorator(func):
            self.router.add_delete(path, self.document(path, "DELETE")(func), **kwargs)
            return func

        return decorator

    def patch(self, path: str, **kwargs):
        """PATCH decorator"""

        def decorator(func):
            self.router.add_patch(path, self.document(path, "PATCH")(func), **kwargs)
            return func

        return decorator

    def head(self, path: str, **kwargs):
        """HEAD decorator"""

        def decorator(func):
            self.router.add_head(path, self.document(path, "HEAD")(func), **kwargs)
            return func

        return decorator

    def options(self, path: str, **kwargs):
        """OPTIONS decorator"""

        def decorator(func):
            self.router.add_options(
                path, self.document(path, "OPTIONS")(func), **kwargs
            )
            return func

        return decorator

    def on_event(self, event: str):
        """On event handler"""

        def decorator(func):
            if event not in ("startup", "shutdown"):
                raise ValueError("Event must be startup or shutdown")
            elif event == "startup":
                self.on_startup.append(func)
            else:
                self.on_shutdown.append(func)
            return func

        return decorator

    def sse(self, path: str) -> Callable:  # pylint: disable=invalid-name
        """Server-Sent Events decorator"""

        def decorator(func: Callable) -> Callable:
            @wraps(func)
            async def wrapper(request: Request) -> EventSourceResponse:
                async with sse_response(request) as resp:
                    args_to_apply = await load(
                        request, signature(func).parameters.copy()
                    )
                    definitive_args = {}
                    for name, param in signature(func).parameters.items():
                        if param.annotation == EventSourceResponse:
                            definitive_args[name] = resp
                        elif name in args_to_apply:
                            definitive_args[name] = args_to_apply[name]
                            args_to_apply.pop(name)
                        elif param.default is not param.empty:
                            definitive_args[name] = param.default
                        else:
                            raise ValueError(
                                f"Missing parameter {name} for {func.__name__}"
                            )
                    await func(**definitive_args)
                    return resp

            self.router.add_get(path, wrapper)
            return wrapper

        return decorator

    def websocket(self, path: str) -> Callable:  # pylint: disable=invalid-name
        """Websocket decorator"""

        def decorator(func: Callable) -> Callable:
            @wraps(func)
            async def wrapper(request: Request):
                args_to_apply = await load(request, signature(func).parameters.copy())
                ws = WebSocketResponse()
                await ws.prepare(request)
                definitive_args = {}
                for name, param in signature(func).parameters.items():
                    if param.annotation == WebSocketResponse:
                        definitive_args[name] = ws
                    elif name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                        args_to_apply.pop(name)
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                await func(**definitive_args)
                return ws

            self.router.add_get(path, wrapper)
            return wrapper

        return decorator

    def static(self):
        """Static folder creation and serving"""
        try:
            os.makedirs("static", exist_ok=True)
        except OSError:
            pass
        self.router.add_static("/", "static")

        @self.get("/")
        def index():
            return FileResponse("static/index.html")

        return self

    def middleware(self, func: Middleware) -> Middleware:
        @wraps(func)
        @middleware
        async def wrapper(request: Request, handler: Handler) -> Response:
            response = await func(request, handler)
            if isinstance(response, Response):
                return response
            return do_response(response)

        self.middlewares.append(wrapper)
        return wrapper



    def run(self, host:str="0.0.0.0", port:int=8080):
        """Run the server"""
        self.should_restart = False
        self.logger.info("Starting server at http://%s:%s", host, port) 
        self.logger.info("Press Ctrl+C to stop the server")
        self.runner = AppRunner(self)

        # Monitor for file changes in a separate thread
        self.file_change_monitor_thread = threading.Thread(target=self.monitor_files, daemon=True)
        self.file_change_monitor_thread.start()

        while True:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self.runner.setup())        
            site = TCPSite(self.runner, host, port)     
            loop.run_until_complete(site.start())
            self.logger.info("Server started")
            self.logger.info("Swagger UI available at http://%s:%s/docs", host, port)
            self.logger.info("OpenAPI JSON available at http://%s:%s/openapi.json", host, port)
            
            try:
                loop.run_forever()
            except KeyboardInterrupt:
                self.logger.info("Stopping server")
                loop.run_until_complete(self.runner.cleanup())
                loop.run_until_complete(self.shutdown())
                loop.run_until_complete(self.cleanup())
                loop.close()
                self.logger.info("Server stopped")
                break
            finally:
                if self.should_restart:
                    self.logger.info("File change detected. Restarting server...")
                    loop.run_until_complete(self.runner.cleanup())
                    loop.run_until_complete(self.shutdown())
                    loop.run_until_complete(self.cleanup())
                    loop.close()
                    self.should_restart = False

    def monitor_files(self):
        """Monitor for changes in Python files and restart the server if a change is detected"""
        paths_to_track = [os.getcwd()]
        file_modification_times = self.track_files(os.getcwd())
        while True:
            for path in paths_to_track:
                for file_path, last_modified in self.track_files(path).items():
                    if file_path not in file_modification_times:
                        file_modification_times[file_path] = last_modified
                    elif file_modification_times[file_path] != last_modified:
                        self.should_restart = True
                        return
            
    @staticmethod
    def track_files(path):
        """Get the modification times for all Python files in the specified path"""
        file_modification_times = {}
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                if filename.endswith(".py"):
                    file_path = os.path.join(dirpath, filename)
                    file_modification_times[file_path] = os.path.getmtime(file_path)
        return file_modification_times

    def restart(self):
        """Restart the server"""
        self.logger.info("Restarting server")
        loop = asyncio.get_event_loop()
        loop.call_soon_threadsafe(loop.stop)


"""Chat Completions Schemas"""
from typing import List, Literal

from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

Role = Literal["assistant", "user", "system", "function"]
Model = Literal["gpt-4-0613", "gpt-3.5-turbo-16k-0613"]


class Message(BaseModel):
    """Defines a message within a conversation."""

    role: Role = Field(..., description="The role of the message.")
    content: str = Field(..., description="The content of the message.")


class ChatCompletionRequest(BaseModel):
    """Defines a request for a chat completion."""

    model: Model = Field(
        default="gpt-4-0613", description="The model used for the chat completion."
    )
    messages: List[Message] = Field(
        ..., description="The list of messages in the conversation."
    )
    temperature: float = Field(
        default=0.2,
        ge=0,
        le=1,
        description="The temperature for the chat completion. Higher values make output more random.",  # pylint: disable=line-too-long
    )
    max_tokens: int = Field(
        default=1024,
        ge=128,
        le=2048,
        description="The maximum number of tokens in the output.",
    )
    stream: bool = Field(default=False, description="Whether the response is a stream.")


class ChatCompletionUssage(BaseModel):
    """Defines the usage of the tokens for a chat completion."""

    prompt_tokens: int = Field(
        ..., description="The number of tokens used in the prompt."
    )
    completion_tokens: int = Field(
        ..., description="The number of tokens used in the completion."
    )
    total_tokens: int = Field(..., description="The total number of tokens used.")


class ChatCompletionChoice(BaseModel):
    """Defines a choice in a chat completion."""

    index: int = Field(..., description="The index of the choice.")
    message: Message = Field(..., description="The message of the choice.")
    finish_reason: str = Field(..., description="The reason for the finish.")


class ChatCompletionResponse(BaseModel):
    """Defines a response for a chat completion."""

    id: str = Field(..., description="The id of the response.")
    object: str = Field(..., description="The object of the response.")
    created: int = Field(..., description="The time when the response was created.")
    model: Model = Field(..., description="The model used for the chat completion.")
    choices: List[ChatCompletionChoice] = Field(
        ..., description="The list of choices in the response."
    )
    usage: ChatCompletionUssage = Field(
        ..., description="The usage of the tokens in the response."
    )
    stream: bool = Field(default=False, description="Whether the response is a stream.")
